from typing import Callable, Dict, Optional, Union

from confluent_kafka import KafkaError, Message

KafkaHeaders = Dict[str, Optional[Union[str, bytes]]]
ProducerOnDelivery = Callable[[KafkaError, Message], None]



import functools
from typing import Set

from pydantic import BaseModel, PositiveInt, PyObject

from kueue._types import KafkaHeaders, ProducerOnDelivery


def pydantic_singleton(clazz):
    class wrapper(clazz):

        _instance = None
        _initialized = False

        @functools.wraps(clazz.__new__)
        def __new__(cls, *args, **kwargs):
            # Only call the real constructor once.
            if cls._instance is None:
                cls._instance = super().__new__(cls)
            return cls._instance

        @functools.wraps(clazz.__init__)
        def __init__(self, *args, **kwargs):
            # Only initialize the instance once.
            if not self.__class__._initialized:
                super().__init__(*args, **kwargs)
            self.__class__._initialized = True

        @classmethod
        def singleton_reset_(cls):
            cls._instance = None
            cls._initialized = False

    wrapper.__module__ = clazz.__module__
    wrapper.__name__ = clazz.__name__
    if hasattr(clazz, "__qualname__"):
        wrapper.__qualname__ = clazz.__qualname__

    return wrapper


class ProducerDefaults(BaseModel):
    """
    Default arguments to use when calling Producer().produce()
    """

    key: str = None
    headers: KafkaHeaders = None
    on_delivery: ProducerOnDelivery = None
    partition: int = None


class ConsumeDefaults(BaseModel):
    """
    Default arguments to use when calling Consumer().consume()
    """

    timeout: int = 5  # seconds
    prefetch: PositiveInt = 1


@pydantic_singleton
class KueueConfig(BaseModel):
    """
    Main config class for Kueue
    """

    kafka_bootstrap: str  # Kafka bootstrap service

    topics: Set = set()  # Topics used / allowed
    default_topic: str = None  # Default topic to use

    # librdkafka config for consumer/producer
    # https://github.com/edenhill/librdkafka/blob/master/CONFIGURATION.md
    consumer_config: dict = {}
    producer_config: dict = {}

    produce_defaults: ProducerDefaults = ProducerDefaults()
    consume_defaults: ConsumeDefaults = ConsumeDefaults()

    task_executor: PyObject = None
    task_config_cls: PyObject = None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kueue']

package_data = \
{'': ['*']}

install_requires = \
['confluent-kafka>=1.5.0,<2.0.0',
 'pydantic>=1.7.3,<2.0.0',
 'wrapt>=1.12.1,<2.0.0']

setup_kwargs = {
    'name': 'kueue',
    'version': '0.1.2',
    'description': 'Distributed Task Queue - backed by Kafka',
    'long_description': '## Kueue\n![Build](https://github.com/jarojasm95/kueue/workflows/Build/badge.svg)\n\nPython Distributed Task Queue - backed by Kafka\n\n\n## Motivation\nKueue was born out of the lack of a library that leverages [Kafka](https://kafka.apache.org/) for distributed task processing\n\n## Features\n- simple\n- intuitive api\n- extensible\n\n## Code Example\n```python\nimport time\nfrom kueue import task, TaskExecutorConsumer, KueueConfig\n\nKueueConfig(\n    kafka_bootstrap=\'localhost:9092\'\n)\n\n@task(topic="my-topic")\ndef sleepy_task(sleep: int):\n    time.sleep(sleep)\n    print("done sleeping", sleep)\n    return sleep\n\nsleepy_task.enqueue(args=(15,))\n\nconsumer = TaskExecutorConsumer(["my-topic"])\nconsumer.start()\n# prints "done sleeping, 15"\n```\n\n## Installation\n```\npip install kueue\n```\n\n## Development\n\nInstall [poetry](https://python-poetry.org/) and run `poetry install` at the root of the repository. This should create a virtual environment with all the necessary python dependencies.\n\n## Tests\nThe test framework makes heavy use of `pytest` fixtures in order to spin up full integration environment consisting of a kubernetes cluster using [kind](https://kind.sigs.k8s.io/) and [pytest-kind](https://codeberg.org/hjacobs/pytest-kind) and kafka using [strimzi](https://strimzi.io/)\n\n`pytest`\n\n## License\n\nMIT © [Jose Rojas](https://github.com/jarojasm95)\n',
    'author': 'Jose Rojas',
    'author_email': 'jose.rojas95@mail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jarojasm95/kueue',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from types import FunctionType, MethodType
from typing import Type

from confluent_kafka import Producer
from pydantic import BaseModel, Field, PyObject, validator
from pydantic.main import ModelMetaclass
from wrapt import FunctionWrapper

from kueue._types import KafkaHeaders, ProducerOnDelivery
from kueue.config import KueueConfig

DEFAULT_JSON_ENCODERS = {
    FunctionType: lambda v: "{}.{}".format(v.__module__, v.__qualname__),
    MethodType: lambda v: "{}.{}".format(v.__module__, v.__qualname__),
    Type: lambda v: "{}.{}".format(v.__module__, v.__qualname__),
    ModelMetaclass: lambda v: "{}.{}".format(v.__module__, v.__qualname__),
}


def dict_if_not_none(**kwargs):
    return {k: v for (k, v) in kwargs.items() if v is not None}


class TaskExecutor(BaseModel):
    """
    Task execution class, encapsulates a task and the parameters used to run it
    """

    class Config:
        json_encoders = DEFAULT_JSON_ENCODERS

    task: PyObject
    args: tuple
    kwargs: dict
    executor: str = None

    @validator("executor", pre=True, always=True)
    def executor_serialize(cls, v):
        if isinstance(v, str):
            return v
        if v is None:
            v = cls
        serializer = cls.Config.json_encoders.get(type(v))
        return serializer(v)

    @property
    def task_name(self):
        serializer = self.Config.json_encoders.get(type(self.task))
        return serializer(self.task)

    def run(self):
        return self.task(*self.args, **self.kwargs)


class TaskProducer(Producer):
    def __init__(self, **config):
        super().__init__(
            {
                "bootstrap.servers": KueueConfig().kafka_bootstrap,
                **KueueConfig().producer_config,
                **config,
            }
        )


class TaskConfig(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    topic: str = Field(default_factory=lambda: KueueConfig().default_topic)
    key: str = Field(default_factory=lambda: KueueConfig().produce_defaults.key)
    producer: TaskProducer = Field(default_factory=TaskProducer)

    headers: KafkaHeaders = Field(default_factory=lambda: KueueConfig().produce_defaults.headers)
    partition: int = Field(default_factory=lambda: KueueConfig().produce_defaults.partition)
    on_delivery: ProducerOnDelivery = Field(
        default_factory=lambda: KueueConfig().produce_defaults.on_delivery
    )

    @validator("topic", always=True)
    def valid_topic(cls, v):
        if not v:
            raise ValueError(f"Invalid task topic: {v}")
        if KueueConfig().topics and v not in KueueConfig().topics:
            raise ValueError(f"Invalid task topic: {v}")
        return v

    def produce(
        self,
        executor: TaskExecutor,
        *,
        key: str = None,
        partition: int = None,
        on_delivery: ProducerOnDelivery = None,
        timestamp: int = None,
        headers: KafkaHeaders = None,
    ):
        kwargs = dict(  # noqa: C408
            key=key or executor.task_name,
            partition=partition or self.partition,
            on_delivery=on_delivery or self.on_delivery,
            timestamp=timestamp,
            headers=headers or self.headers,
        )
        return self.producer.produce(
            self.topic,
            executor.json(),
            **{k: v for (k, v) in kwargs.items() if v is not None},
        )


class TaskProxy(FunctionWrapper):
    def __init__(
        self,
        wrapped,
        *,
        task_config_cls: Type[TaskConfig],
        task_executor: Type[TaskExecutor],
        **kwargs,
    ):
        super(TaskProxy, self).__init__(wrapped, wrapped, enabled=False)
        task_config_cls = task_config_cls or KueueConfig().task_config_cls or TaskConfig
        task_executor = task_executor or KueueConfig().task_executor or TaskExecutor
        self.task_config: TaskConfig = task_config_cls(**kwargs)
        self.task_executor: Type[TaskExecutor] = task_executor

    @property
    def name(self):
        return f"{self.__wrapped__.__module__}.{self.__wrapped__.__qualname__}"

    def enqueue(
        self, *, args: tuple = (), kwargs: dict = {}, task_extra: dict = {}, **producer_options
    ):
        return self.task_config.produce(
            self.task_executor(
                task=self.__wrapped__,
                args=args,
                kwargs=kwargs,
                **task_extra,
            ),
            **producer_options,
        )

    def __repr__(self):
        return f"<@task: {self.name}>"

    __str__ = __repr__


def task(
    func=None,
    task_config_cls: Type[TaskConfig] = None,
    task_executor: Type[TaskExecutor] = None,
    **task_config,
):
    def _wrapper(func):
        return TaskProxy(
            func, task_config_cls=task_config_cls, task_executor=task_executor, **task_config
        )

    if func is None:
        return _wrapper
    return _wrapper(func)



import logging
import signal
import socket
from contextlib import contextmanager
from enum import Enum
from typing import List, Optional

from confluent_kafka import (
    TIMESTAMP_CREATE_TIME,
    TIMESTAMP_LOG_APPEND_TIME,
    TIMESTAMP_NOT_AVAILABLE,
    Consumer,
    KafkaError,
    KafkaException,
    Message,
    ThrottleEvent,
    TopicPartition,
)
from pydantic import BaseModel, Json, PyObject, validator

from kueue.config import KueueConfig
from kueue.task import TaskExecutor


class TimestampType(Enum):

    NotAvailable = TIMESTAMP_NOT_AVAILABLE
    Broker = TIMESTAMP_LOG_APPEND_TIME
    Producer = TIMESTAMP_CREATE_TIME


class TaskMessage(BaseModel):

    topic: str
    partition: int
    offset: int
    key: Optional[str] = None
    value: Json
    timestamp_type: TimestampType
    timestamp: Optional[int]

    @validator("timestamp")
    def timestamp_to_seconds(cls, v, values):
        if values.get("timestamp_type") == TimestampType.NotAvailable:
            return None
        return v / 1000.0

    def executor(self) -> TaskExecutor:
        return PyObject.validate(self.value["executor"]).parse_obj(self.value)

    @classmethod
    def parse_kafka_message(cls, message: Message) -> "TaskMessage":
        ts_type, ts = message.timestamp()
        return cls(
            topic=message.topic(),
            partition=message.partition(),
            offset=message.offset(),
            key=message.key(),
            value=message.value(),
            timestamp=ts,
            timestamp_type=ts_type,
            size=len(message),
        )


class ConsumerBase:
    def __init__(self, **config):
        name = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
        self.logger = logging.getLogger(name)
        self.kafka = Consumer(
            {
                "bootstrap.servers": KueueConfig().kafka_bootstrap,
                "group.id": name,
                "client.rack": socket.gethostname(),
                "enable.auto.commit": False,
                "on_commit": self.on_commit,
                "error_cb": self.on_internal_error,
                "throttle_cb": self.on_throttle,
                "auto.offset.reset": "earliest",
                "api.version.request": True,
                **KueueConfig().consumer_config,
                **config,
            }
        )
        self._exit = False

    def on_assign(self, _, partitions: List[TopicPartition]):
        self.logger.info("on_assign: %s", partitions)

    def on_revoke(self, _, partitions: List[TopicPartition]):
        self.logger.info("on_revoke: %s", partitions)

    def on_exit(self):
        self.logger.info("on_exit")

    def on_throttle(self, event: ThrottleEvent):
        self.logger.warning("on_throttle")

    def on_stats(self, stats: str):
        self.logger.info("on_stats: %s", stats)

    def on_commit(self, error: Optional[KafkaError], partitions: List[TopicPartition]):
        self.logger.info("on_commit: %s | %s", error, partitions)

    def on_timeout(self):
        self.logger.warning("on_timeout")

    def on_internal_error(self, error: KafkaError, message: Message = None):
        self.logger.exception("on_internal_error: %s | %s", error, message)

    def on_eof(self):
        self.logger.warning("on_eof")

    def consume(self) -> List[Message]:
        try:
            messages: List[Message] = self.kafka.consume(
                num_messages=KueueConfig().consume_defaults.prefetch,
                timeout=KueueConfig().consume_defaults.timeout,
            )
        except KafkaError as e:
            self.on_internal_error(e)
            return []
        except KafkaException as e:
            self.on_internal_error(e.args[0])
            return []
        if not messages:
            self.on_timeout()
            return []
        errors = list(filter(lambda msg: msg.error(), messages))
        messages = list(filter(lambda msg: not msg.error(), messages))
        for error_msg in errors:
            if error_msg.error().code() == KafkaError._PARTITION_EOF:
                self.on_eof(error_msg)
            else:
                self.on_internal_error(error_msg.error(), error_msg)
        return messages

    def _iterator(self):
        while not self._exit:
            messages = self.consume()
            for message in messages:
                yield message

    def _setup_signal_handling(self):
        signal.signal(signal.SIGTERM, self.stop)

    def stop(self, signal=None, frame=None):
        self._exit = True
        self.on_exit()

    @contextmanager
    def stream(self):
        try:
            yield self._iterator()
        finally:
            self.kafka.close()


class TaskExecutorConsumer(ConsumerBase):
    """
    Consumer class for task execution
    """

    def __init__(self, topics: List[str], **config):
        super().__init__(**config)
        self.kafka.subscribe(topics, on_assign=self.on_assign, on_revoke=self.on_revoke)
        self._setup_signal_handling()

    def on_message(self, message: TaskMessage):
        return message.executor().run()

    def on_success(self, message: TaskMessage, result):
        self.logger.info("on_success: %s | %s", message, result)

    def on_error(self, message: TaskMessage):
        self.logger.exception("on_error: %s", message)

    def on_commit(self, error: Optional[KafkaError], partitions: List[TopicPartition]):
        self.logger.info("on_commit: %s | %s", error, partitions)

    def dispatch_message(self, message: Message):
        parsed = TaskMessage.parse_kafka_message(message)
        try:
            result = self.on_message(parsed)
        except Exception:
            try:
                self.on_error(parsed)
            except Exception:
                self.logger.exception("error calling on_error")
        else:
            try:
                self.on_success(parsed, result)
            except Exception:
                self.logger.exception("error calling on_success")

        try:
            commits: List[TopicPartition] = self.kafka.commit(message=message, asynchronous=False)
        except KafkaException as e:
            self.on_commit(e.args[0], [])
        else:
            self.on_commit(None, commits)

    def start(self):
        """
        Start message consumption
        Should be used when the consumer is defined as a subclass of `Consumer`
        or if callbacks are configured using the `*_handler` decorators
        """
        with self.stream() as message_stream:
            for message in message_stream:
                self.dispatch_message(message)
